from logging import getLogger
from uuid import UUID
from typing import Union, Optional, List

from gemd.entity.link_by_uid import LinkByUID
from gemd.enumeration.base_enumeration import BaseEnumeration

from citrine.seeding.find_or_create import find_collection, create_or_update
from citrine.jobs.waiting import wait_while_validating
from citrine.informatics.data_sources import GemTableDataSource
from citrine.informatics.executions import DesignExecution
from citrine.informatics.design_candidate import DesignCandidate
from citrine.informatics.design_spaces import DesignSpace
from citrine.informatics.predictor_evaluator import PredictorEvaluator
from citrine.informatics.objectives import Objective
from citrine.informatics.scores import Score
from citrine.informatics.workflows import DesignWorkflow, PredictorEvaluationWorkflow

from citrine.resources.gemtables import GemTable
from citrine.resources.material_run import MaterialRun
from citrine.resources.predictor import Predictor, AutoConfigureMode
from citrine.resources.project import Project
from citrine.resources.table_config import TableConfig, TableBuildAlgorithm

from citrine.builders.scores import create_default_score

logger = getLogger(__name__)


class AutoConfigureStatus(BaseEnumeration):
    """[ALPHA] The current status of the AutoConfigureWorkflow.

    * START is the initial status before auto-configuration
    * TABLE_BUILD is the status after creating a default GEM table
    * PREDICTOR_CREATED is the status after creating a default predictor
    * PREDICTOR_INVALID is the status if predictor validation fails
    * PEW_CREATED is the status after creating a predictor evaluation workflow
    * PEW_FAILED is the status if predictor evaluation workflow validation fails
    * DESIGN_SPACE_CREATED is the status after design space creation/registration
    * DESIGN_SPACE_INVALID is the status if design space validation fails
    * DESIGN_WORKFLOW_CREATED is the status after design workflow creation
    * DESIGN_WORKFlOW_FAILED is the status if design workflow validation fails
    """

    START = 'START'
    TABLE_CREATED = 'TABLE CREATED'
    PREDICTOR_CREATED = 'PREDICTOR CREATED'
    PREDICTOR_INVALID = 'PREDICTOR INVALID'
    PEW_CREATED = 'PREDICTOR EVALUATION WORKFLOW CREATED'
    PEW_FAILED = 'PREDICTOR EVALUATION WORKFLOW FAILED'
    DESIGN_SPACE_CREATED = 'DESIGN SPACE CREATED'
    DESIGN_SPACE_INVALID = 'DESIGN SPACE INVALID'
    DESIGN_WORKFLOW_CREATED = 'DESIGN WORKFLOW CREATED'
    DESIGN_WORKFLOW_FAILED = 'DESIGN WORKFLOW FAILED'


class AutoConfigureWorkflow():
    """[ALPHA] Helper class for configuring and storing default assets on the Citrine Platform.

    Defines methods for creating a default GEM table, predictor, predictor evaluations,
    design space, and design workflow in a linear fashion,
    starting from a provided material/table/predictor.

    All assets that are registered to the Citrine Platform
    during the auto configuration steps are stored as members of this class.
    In case this method fails during asset validation,
    the previously configured items are accessible
    locally and on the Citrine Platform.

    Initializing an AutoConfigureWorkflow will search for assets
    on the Citrine Platform based on the specified name,
    and re-use/update these items during subsequent runs.
    Currently, assets are searched for by name and type in the project,
    and duplicately named assets will result in an error being thrown.

    Example usage:
    ```
    # Initialize score and workflow
    score = LIScore(objectives=[ScalarMaxObjective(descriptor_key='Desc 1'), baselines=[1.0])
    auto_config = AutoConfigureWorkflow(project=project, name='My Project Name')

    # GEM table -> Predictor -> PEW/PEE -> Design Space -> Design Workflow -> Design Execution
    auto_config.from_material(
        material=sample_material,
        score=score,
        mode=AutoConfigureMode.PLAIN,
        print_status_info=True
    )

    # Check the most recent status
    print(f"Workflow ended with status: {auto_config.status}")

    # Retrieve some of the results
    table = auto_config.table
    predictor = auto_config.predictor
    design_space = auto_config.design_space
    workflow = auto_config.design_workflow
    ```

    Parameters
    ----------
    project: Project
        Project to use when accessing the Citrine Platform.
    name: str
        Name to affix to auto-configured assets on the Citrine Platform.
        This name is used as an identifier for finding and re-using
        assets related to a given workflow.

    """

    def __init__(self, *, project: Project, name: str):
        self._project = project
        self._name = name
        self._status = AutoConfigureStatus.START
        self._status_info = []

        # Blank initialize assets
        self._table_config = None
        self._table = None
        self._predictor = None
        self._predictor_evaluation_workflow = None
        self._design_space = None
        self._design_workflow = None
        self._design_execution = None

        separator = ": " if self.name else ""
        self._default_asset_names = {
            "TABLE_CONFIG": f"{self.name}{separator}Auto Configure GEM Table",
            "PREDICTOR": f"{self.name}{separator}Auto Configure Predictor",
            "PEW": f"{self.name}{separator}Auto Configure PEW",
            "DESIGN_SPACE": f"{self.name}{separator}Auto Configure Design Space",
            "DESIGN_WORKFLOW": f"{self.name}{separator}Auto Configure Design Workflow"
        }

        # Search project for current status
        self.update()

    @staticmethod
    def _print_status(msg: str):
        logger.info(f"AutoConfigureWorkflow: {msg}")

    @property
    def project(self) -> Project:
        """Get the project used when automatically configuring assets."""
        return self._project

    @property
    def name(self) -> str:
        """Get the naming label for assets configured by this object."""
        return self._name

    @property
    def status(self) -> str:
        """Get the most recent status of the auto-configure workflow."""
        return self._status.value

    @property
    def status_info(self) -> List[str]:
        """Get the most recent status info from asset configuration/validation."""
        return self._status_info

    @property
    def table_config(self) -> Optional[TableConfig]:
        """Get the table config associated with this workflow."""
        return self._table_config

    @property
    def table(self) -> Optional[GemTable]:
        """Get the GEM table associated with this workflow."""
        return self._table

    @property
    def predictor(self) -> Optional[Predictor]:
        """Get the predictor associated with this workflow."""
        return self._predictor

    @property
    def predictor_evaluation_workflow(self) -> Optional[PredictorEvaluationWorkflow]:
        """Get the predictor evaluation workflow associated with this workflow."""
        return self._predictor_evaluation_workflow

    @property
    def design_space(self) -> Optional[DesignSpace]:
        """Get the design space associated with this workflow."""
        return self._design_space

    @property
    def design_workflow(self) -> Optional[DesignWorkflow]:
        """Get the design workflow associated with this workflow."""
        return self._design_workflow

    @property
    def design_execution(self) -> Optional[DesignExecution]:
        """Get the most recent design execution from this workflow."""
        return self._design_execution

    @property
    def score(self) -> Optional[Score]:
        """Get the most recent score executed by this workflow."""
        de = self.design_execution
        return de.score if de is not None else None

    @property
    def candidates(self) -> List[DesignCandidate]:
        """Get the candidate list from the most recent design execution."""
        de = self.design_execution
        return list(de.candidates()) if de is not None else []

    @property
    def assets(self):
        """Get all assets configured by this object."""
        initial_assets = [
            self.table_config, self.table, self.predictor,
            self.predictor_evaluation_workflow,
            self.design_space, self.design_workflow,
        ]
        return [asset for asset in initial_assets if asset is not None]

    def update(self):
        """Search for existing assets matching the workflow name and update its status."""
        self._update_assets()
        self._update_status()

    def _update_assets(self):
        """Find and store all assets on the platform matching the workflow name."""
        # Table config
        if self.table_config is None:
            self._table_config = find_collection(
                collection=self.project.table_configs,
                name=self._default_asset_names["TABLE_CONFIG"]
            )
        else:
            self._table_config = self.project.table_configs.get(self.table_config.uid)
            logger.info("Found existing: {}".format(self.table_config))

        # Table
        if self.table is None:
            if self.table_config is not None:
                self._table = next(self.project.tables.list_by_config(self.table_config.uid), None)
                if self.table is not None:
                    logger.info("Found existing: {}".format(self.table))
        else:
            self._table = self.project.tables.get(self.table.uid)
            logger.info("Found existing: {}".format(self.table))

        # Predictor
        if self.predictor is None:
            self._predictor = find_collection(
                collection=self.project.predictors,
                name=self._default_asset_names["PREDICTOR"]
            )
        else:
            self._predictor = self.project.predictors.get(self.predictor.uid)
            logger.info("Found existing: {}".format(self.predictor))

        # PEW
        if self.predictor_evaluation_workflow is None:
            self._predictor_evaluation_workflow = find_collection(
                collection=self.project.predictor_evaluation_workflows,
                name=self._default_asset_names["PEW"]
            )
        else:
            self._predictor_evaluation_workflow = self.project.predictor_evaluation_workflows.get(
                self.predictor_evaluation_workflow.uid
            )
            logger.info("Found existing: {}".format(self.predictor_evaluation_workflow))

        # Design space
        if self.design_space is None:
            self._design_space = find_collection(
                collection=self.project.design_spaces,
                name=self._default_asset_names["DESIGN_SPACE"]
            )
        else:
            self._design_space = self.project.design_spaces.get(self.design_space.uid)
            logger.info("Found existing: {}".format(self.design_space))

        # Design workflow
        if self.design_workflow is None:
            self._design_workflow = find_collection(
                collection=self.project.design_workflows,
                name=self._default_asset_names["DESIGN_WORKFLOW"]
            )
        else:
            self._design_workflow = self.project.design_workflows.get(self.design_workflow.uid)
            logger.info("Found existing: {}".format(self.design_workflow))

    def _update_status(self):
        """Update status info based on currently stored assets."""
        # Work backwards from end of workflow, return early when possible
        if self.design_workflow is not None:
            self._status_info = self.design_workflow.status_info
            if self.design_workflow.failed():
                self._status = AutoConfigureStatus.DESIGN_WORKFLOW_FAILED
            else:
                self._status = AutoConfigureStatus.DESIGN_WORKFLOW_CREATED
            return

        if self.design_space is not None:
            self._status_info = [detail.msg for detail in self.predictor.status_detail]
            if self.design_space.failed():
                self._status = AutoConfigureStatus.DESIGN_SPACE_INVALID
            else:
                self._status = AutoConfigureStatus.DESIGN_SPACE_CREATED
            return

        if self.predictor_evaluation_workflow is not None:
            self._status_info = self.predictor_evaluation_workflow
            if self.predictor_evaluation_workflow.failed():
                self._status = AutoConfigureStatus.PEW_FAILED
            else:
                self._status = AutoConfigureStatus.PEW_CREATED
            return

        if self.predictor is not None:
            self._status_info = [detail.msg for detail in self.predictor.status_detail]
            if self.predictor.failed():
                self._status = AutoConfigureStatus.PREDICTOR_INVALID
            else:
                self._status = AutoConfigureStatus.PREDICTOR_CREATED
            return

        if self.table is not None:
            self._status = AutoConfigureStatus.TABLE_CREATED
            return

        self._status = AutoConfigureStatus.START

    def execute(self, *, score: Union[Score, Objective]) -> DesignExecution:
        """[ALPHA] Execute a design execution using the provided score/objective.

        A design workflow must be present after auto-configuration to execute the score.
        This method can be called after modifying auto-configured assets
        to create a new, custom design execution.

        Parameters
        ----------
        score: Union[Score, Objective]
            A score or objective used to rank candidates during design execution.
            If an objective is passed, a default score is constructed
            by reading the GEM table associated with this workflow
            to extract baseline information.

        Returns
        ----------
        DesignExecution
            The design execution triggered with the provided score/objective.

        """
        if self.design_workflow is None:
            raise ValueError("Design workflow is missing, cannot execute score.")

        if isinstance(score, Objective):
            score = create_default_score(objectives=score, project=self.project, table=self.table)

        self._design_execution = self.design_workflow.design_executions.trigger(score)
        return self.design_execution

    def from_material(
            self,
            *,
            material: Union[str, UUID, LinkByUID, MaterialRun],
            evaluator: Optional[PredictorEvaluator] = None,
            design_space: Optional[DesignSpace] = None,
            score: Optional[Union[Score, Objective]] = None,
            mode: AutoConfigureMode = AutoConfigureMode.PLAIN,
            print_status_info: bool = False,
            prefer_valid: bool = True,
    ):
        """[ALPHA] Auto configure platform assets from material history to design workflow.

        Given a material on the Citrine Platform,
        configures a default GEM table, predictor, design space, and design workflow.

        An optional evaluator, design_space, and score can be provided:
            evaluator    -> A PredictorEvaluator to be used in place of the default
            design_space -> A DesignSpace to be used in place of the default
            score        -> Triggers a design execution from the auto-configured design workflow

        Parameters
        ----------
        material: Union[str, UUID, LinkByUID, MaterialRun]
            A representation of the material to configure a
            default table, predictor, and design space from.
        evaluator: Optional[PredictorEvaluator]
            A PredictorEvaluator to use in place of the default.
            Must contain responses matching outputs of the default predictor.
            Default: None
        design_space: Optional[DesignSpace]
            A DesignSpace object to use in place of the default.
            If not registered already, will be registered during configuration.
            Default: None
        score: Optional[Union[Score, Objective]]
            A score or objective used to rank candidates during design execution.
            If only an objective is passed, a default `LIScore` is constructed
            by reading the GEM table associated with this workflow
            to extract baseline scoring information.
            Default: None
        mode: AutoConfigureMode
            The method to be used in the automatic table and predictor configuration.
            Default: AutoConfigureMode.PLAIN
        print_status_info: bool
            Whether to print the status info during validation of assets.
            Default: False
        prefer_valid: Boolean
            If True, enables filtering of sparse descriptors and trimming of
            excess graph components in attempt to return a default configuration
            that will pass validation.
            Default: True.

        """
        if not isinstance(mode, AutoConfigureMode):
            raise TypeError('mode must be an option from AutoConfigureMode')

        # Run table build stage
        self._table_build_stage(material=material, mode=mode)

        # Finish workflow from table stage
        self.from_table(
            table=self.table, score=score, evaluator=evaluator,
            design_space=design_space, mode=mode, print_status_info=print_status_info,
            prefer_valid=prefer_valid
        )

    def from_table(
            self,
            *,
            table: GemTable,
            evaluator: Optional[PredictorEvaluator] = None,
            design_space: Optional[DesignSpace] = None,
            score: Optional[Union[Score, Objective]] = None,
            mode: AutoConfigureMode = AutoConfigureMode.PLAIN,
            print_status_info: bool = False,
            prefer_valid: bool = True,
    ):
        """[ALPHA] Auto configure platform assets from GEM table to design workflow.

        Given a GEM table on the Citrine Platform,
        creates a default predictor, design space, and design workflow.

        An optional evaluator, design_space, and score can be provided:
            evaluator    -> A PredictorEvaluator to be used in place of the default
            design_space -> A DesignSpace to be used in place of the default
            score        -> Triggers a design execution from the auto-configured design workflow

        Parameters
        ----------
        table: Table
            A GEM table to configure a default predictor,
            design space, and design workflow from.
        evaluator: Optional[PredictorEvaluator]
            A PredictorEvaluator to use in place of the default.
            Must contain responses matching outputs of the default predictor.
            Default: None
        design_space: Optional[DesignSpace]
            A DesignSpace object to use in place of the default.
            If not registered already, will be registered during configuration.
            Default: None
        score: Optional[Union[Score, Objective]]
            A score or objective used to rank candidates during design execution.
            If only an objective is passed, a default `LIScore` is constructed
            by reading the GEM table associated with this workflow
            to extract baseline scoring information.
            Default: None
        mode: AutoConfigureMode
            The method to be used in the automatic predictor configuration.
            Default: AutoConfigureMode.PLAIN
        print_status_info: bool
            Whether to print the status info during validation of assets.
            Default: False
        prefer_valid: Boolean
            If True, enables filtering of sparse descriptors and trimming of
            excess graph components in attempt to return a default configuration
            that will pass validation.
            Default: True.

        """
        if not isinstance(mode, AutoConfigureMode):
            raise TypeError('mode must be an option from AutoConfigureMode')

        self._table = table  # Save here, but cannot force rename w/o config

        # Get default predictor, pass to next stage for registration
        data_source = GemTableDataSource(table_id=table.uid, table_version=table.version)
        predictor = self.project.predictors.create_default(
            training_data=data_source, pattern=mode.value,  # Uses same string pattern
            prefer_valid=prefer_valid
        )

        # Finish workflow from predictor stage
        self.from_predictor(
            predictor=predictor, score=score, evaluator=evaluator,
            design_space=design_space, print_status_info=print_status_info
        )

    def from_predictor(
            self,
            *,
            predictor: Predictor,
            evaluator: Optional[PredictorEvaluator] = None,
            design_space: Optional[DesignSpace] = None,
            score: Optional[Union[Score, Objective]] = None,
            print_status_info: bool = False
    ):
        """[ALPHA] Auto configure platform assets from predictor to design workflow.

        Given a predictor on the Citrine Platform,
        creates a default design space and design workflow.

        An optional evaluator, design_space, and score can be provided:
            evaluator    -> A PredictorEvaluator to be used in place of the default
            design_space -> A DesignSpace to be used in place of the default
            score        -> Triggers a design execution from the auto-configured design workflow

        Parameters
        ----------
        predictor: Predictor
            A registered predictor to configure a default design space and design workflow from.
        evaluator: Optional[PredictorEvaluator]
            A PredictorEvaluator to use in place of the default.
            Must contain responses matching outputs of the default predictor.
            Default: None
        design_space: Optional[DesignSpace]
            A DesignSpace object to use in place of the default.
            If not registered already, will be registered during configuration.
            Default: None
        score: Optional[Union[Score, Objective]]
            A score or objective used to rank candidates during design execution.
            If only an objective is passed, a default `LIScore` is constructed
            by reading the GEM table associated with this workflow
            to extract baseline scoring information.
            Default: None
        print_status_info: bool
            Whether to print the status info during validation of assets.
            Default: False

        """
        # Run the predictor registration stage
        self._predictor_registration_stage(
            predictor=predictor,
            print_status_info=print_status_info
        )

        # Run the predictor evaluation stage
        self._predictor_evaluation_stage(
            predictor=self.predictor,
            evaluator=evaluator,
            print_status_info=print_status_info
        )

        # Run the design space stage
        self._design_space_build_stage(
            predictor=self.predictor,
            design_space=design_space,
            print_status_info=print_status_info
        )

        # Run the design workflow build stage, trigger score if provided
        self._design_workflow_build_stage(
            predictor=self.predictor,
            design_space=self.design_space,  # Use the object from previous stage
            score=score,
            print_status_info=print_status_info
        )

    def _table_build_stage(
            self,
            *,
            material: Union[str, UUID, LinkByUID, MaterialRun],
            mode: AutoConfigureMode
    ):
        """Create a default GEM table from a material."""
        self._print_status("Configuring GEM table...")

        # TODO: package-wide formatting enum for method
        table_algorithm_map = {
            'PLAIN': TableBuildAlgorithm.SINGLE_ROW,
            'FORMULATION': TableBuildAlgorithm.FORMULATIONS
        }
        table_algorithm = table_algorithm_map[mode.value]

        table_config, _ = self.project.table_configs.default_for_material(
            material=material,
            name=self._default_asset_names["TABLE_CONFIG"],
            algorithm=table_algorithm
        )
        table_config = create_or_update(
            collection=self.project.table_configs,
            resource=table_config
        )

        self._table_config = table_config
        self._table = self.project.tables.build_from_config(self.table_config)
        self._status = AutoConfigureStatus.TABLE_CREATED

    def _predictor_registration_stage(
            self,
            *,
            predictor: Predictor,
            print_status_info: bool
    ):
        """Register and validate a provided predictor."""
        self._print_status("Configuring predictor...")

        predictor.name = self._default_asset_names["PREDICTOR"]
        predictor = create_or_update(
            collection=self.project.predictors,
            resource=predictor
        )
        predictor = wait_while_validating(
            collection=self.project.predictors,
            module=predictor,
            print_status_info=print_status_info
        )

        self._predictor = predictor
        self._status = AutoConfigureStatus.PREDICTOR_CREATED
        self._status_info = [detail.msg for detail in predictor.status_detail]

        if predictor.failed():
            self._status = AutoConfigureStatus.PREDICTOR_INVALID
            raise RuntimeError("Predictor is invalid,"
                               " cannot proceed to design space configuration.")

    def _predictor_evaluation_stage(
            self,
            *,
            predictor: Predictor,
            evaluator: Optional[PredictorEvaluator],
            print_status_info: bool
    ):
        """Evaluate the predictor with the specified evaluator, or create a default workflow."""
        self._print_status("Configuring predictor evaluation...")

        if evaluator is None:
            # No evaluator, so create a default PEW for the predictor
            pew = self.project.predictor_evaluation_workflows.create_default(
                predictor_id=predictor.uid
            )
        else:
            # We got an evaluator, so make a new PEW and register it manually
            pew = PredictorEvaluationWorkflow(
                name=self._default_asset_names["PEW"],
                evaluators=[evaluator]
            )
            pew = create_or_update(
                collection=self.project.predictor_evaluation_workflows,
                resource=pew
            )

        pew = wait_while_validating(
            collection=self.project.predictor_evaluation_workflows,
            module=pew,
            print_status_info=print_status_info
        )

        self._predictor_evaluation_workflow = pew
        self._status = AutoConfigureStatus.PEW_CREATED
        self._status_info = pew.status_info

        if pew.failed():
            # Can proceed without raising error, but can't get PEE
            self._status = AutoConfigureStatus.PEW_FAILED
            logger.warning(
                "Predictor evaluation workflow failed -- unable to configure execution."
            )
        elif evaluator is not None:
            # Manually trigger execution
            pew.executions.trigger(predictor.uid)

    def _design_space_build_stage(
            self,
            *,
            predictor: Predictor,
            design_space: Optional[DesignSpace],
            print_status_info: bool
    ):
        """Create a design space from a specified predictor, or use the provided one."""
        self._print_status("Configuring design space...")

        if design_space is None:
            design_space = self.project.design_spaces.create_default(predictor_id=predictor.uid)

        design_space.name = self._default_asset_names["DESIGN_SPACE"]
        design_space = create_or_update(
            collection=self.project.design_spaces,
            resource=design_space
        )
        design_space = wait_while_validating(
            collection=self.project.design_spaces, module=design_space,
            print_status_info=print_status_info
        )

        self._design_space = design_space
        self._status = AutoConfigureStatus.DESIGN_SPACE_CREATED
        self._status_info = [detail.msg for detail in design_space.status_detail]

        if design_space.failed():
            self._status = AutoConfigureStatus.DESIGN_SPACE_INVALID
            raise RuntimeError("Design space is invalid,"
                               " cannot proceed to design workflow configuration.")

    def _design_workflow_build_stage(
            self,
            *,
            predictor: Predictor,
            design_space: DesignSpace,
            score: Optional[Union[Score, Objective]],
            print_status_info: bool
    ):
        """Create a design workflow, and optionally trigger a design execution."""
        self._print_status("Configuring design workflow...")

        processor_id = self.design_workflow.processor_id if self.design_workflow else None
        workflow = DesignWorkflow(
            name=self._default_asset_names["DESIGN_WORKFLOW"],
            predictor_id=predictor.uid,
            design_space_id=design_space.uid,
            processor_id=processor_id
        )
        workflow = create_or_update(
            collection=self.project.design_workflows,
            resource=workflow
        )
        workflow = wait_while_validating(
            collection=self.project.design_workflows, module=workflow,
            print_status_info=print_status_info
        )

        self._design_workflow = workflow
        self._status = AutoConfigureStatus.DESIGN_WORKFLOW_CREATED
        self._status_info = workflow.status_info

        if workflow.failed():
            self._status = AutoConfigureStatus.DESIGN_WORKFLOW_FAILED
            if score is None:
                self._print_status("No score provided to trigger design execution -- finished.")
            else:
                raise RuntimeError("Design workflow validation failed,"
                                   " cannot trigger design execution.")
        elif score is not None:
            self._print_status("Triggering design execution...")
            return self.execute(score=score)



from typing import List, Optional

from citrine._rest.engine_resource import VersionedEngineResource
from citrine._serialization import properties as _properties
from citrine.informatics.descriptors import ChemicalFormulaDescriptor, Descriptor
from citrine.informatics.predictors import Predictor

__all__ = ['ChemicalFormulaFeaturizer']


class ChemicalFormulaFeaturizer(VersionedEngineResource['ChemicalFormulaFeaturizer'], Predictor):
    """
    A featurizer for chemical formulae. Inspired by Magpie.

    The ChemicalFormulaFeaturizer computes a configurable set of features on chemical formula data.
    The features are functions of element-level properties, which are inspired by
    `Magpie <https://bitbucket.org/wolverton/magpie/src/master/>`_. The features are configured
    using the ``features`` and ``excludes`` arguments, which accept either feature names or
    predefined aliases. Many features are stoichiometrically weighted generalized means of
    element-level properties. How to compute the mean is configured using the ``powers`` argument.

    The default is the "standard" alias, corresponding to features that are intuitive and
    often correlate with properties of interest. Other aliases are "physical," "electronic,"
    and "periodicTable."

    The following features are weighted means of simple elemental properties.

    - "Pauling electronegativity": standard, electronic
    - "Number of d valence electrons": standard, electronic
    - "Number of unfilled f valence electrons": standard, electronic
    - "Number of f valence electrons": standard, electronic
    - "Number of unfilled p valence electrons": standard, electronic
    - "Number of p valence electrons": standard, electronic
    - "Number of unfilled s valence electrons": standard, electronic
    - "Number of s valence electrons": standard, electronic
    - "Total number of unfilled valence electrons": standard, electronic
    - "Total number of valence electrons": standard, electronic
    - "Elemental work function": standard, electronic
    - "Elemental polarizability": standard, electronic
    - "Radius of d orbitals": standard, electronic
    - "Radius of s orbitals": standard, electronic
    - "Radius of p orbitals": standard, electronic
    - "Elemental magnetic moment": standard, electronic
    - "Elemental atomic volume": standard, electronic, physical
    - "Elemental electron density": standard, electronic
    - "Mendeleev number": standard, periodicTable
    - "Row in periodic table": standard, periodicTable
    - "Elemental bulk modulus": standard, physical
    - "Elemental density": standard, physical
    - "Elemental melting temperature": standard, physical
    - "Elemental crystal structure (space group)": standard, electronic, physical
    - "AtomicVolume": electronic, physical
    - "Number": periodicTable
    - "CovalentRadius": electronic, physical
    - "DipolePolarizability": electronic
    - "ElectronAffinity": electronic
    - "FirstIonizationEnergy": electronic
    - "GSbandgap": electronic
    - "GSenergy_pa": electronic
    - "GSestBCClatcnt": electronic, physical
    - "GSvolume_pa": electronic, physical
    - "MiracleRadius": electronic, physical
    - "NdUnfilled": electronic
    - "ZungerPP-r_pi": electronic
    - "AtomicWeight": physical, periodicTable
    - "Column in periodic table": periodicTable
    - "IsAlkali": periodicTable
    - "IsDBlock": periodicTable
    - "IsFBlock": periodicTable
    - "IsMetal": periodicTable
    - "IsNonmetal": periodicTable
    - "BoilingT": physical
    - "FusionEnthalpy": physical
    - "HeatCapacityMass": physical
    - "HeatCapacityMolar": physical
    - "HeatFusion": physical
    - "ShearModulus": physical
    - "ValenceZeff": electronic, physical

    The following features are weighted means of more complex elemental properties.

    - "Packing density": standard, physical
    - "Liquid range": standard, physical
    - "Non-dimensional liquid range": standard, physical
    - "Liquid ratio": standard, physical
    - "Elastic Poisson Ratio": standard, physical
    - "DFT energy density": standard, electronic, physical
    - "Interatomic distance": standard, physical
    - "Ionization Affinity Ratio": standard, electronic
    - "Ratio of Electron Affinity to Electronegativity": standard, electronic
    - "Trouton's Ratio": standard, physical
    - "Miracle Ratio": standard, electronic
    - "DFT volume ratio": standard, physical
    - "Mulliken electronegativity": standard, electronic
    - "Modulii sum": standard, physical
    - "Zunger Pseudopotential radius ratio": standard, electronic
    - "BCC Efficiency": standard, physical
    - "Non-dimensional heat of fusion": standard, physical
    - "Non-dimensional band gap": standard, electronic
    - "Conduction ionization energy": standard, electronic
    - "Valence electron density": standard, electronic
    - "Non-dimensional work function": standard, electronic
    - "Shear Modulus Melting Temp Product": standard, physical

    The following features are not weighted means. Their values do not depend on ``powers``.

    - "Maximum electronegativity difference": standard, electronic
    - "Maximum radius difference": standard, electronic, physical
    - "Maximum radius ratio": standard, electronic, physical
    - "Min atomic radius plus max electronegativity difference": standard, electronic, physical
    - "Number of elements"
    - "Minimum atomic fraction"
    - "Maximum atomic fraction"
    - "Minimum weight fraction": standard, periodicTable
    - "Maximum weight fraction": standard, periodicTable
    - "Formula weight": standard, physical

    Parameters
    ----------
    input_descriptor: ChemicalFormulaDescriptor
        the descriptor to featurize
    features: Optional[List[str]]
        The list of features to compute, either by name or by group alias. Default is "standard."
    excludes: Optional[List[str]]
        The list of features to exclude, either by name or by group alias. Default is none.
        The final set of features generated by the predictor is set(features) - set(excludes).
    powers: Optional[List[int]]
        The list of powers to use when computing generalized weighted means of element properties.
        p=1 corresponds to the ordinary mean, p=2 is the root mean square, etc.

    """

    input_descriptor = _properties.Object(Descriptor, 'data.instance.input')
    features = _properties.List(_properties.String, 'data.instance.features')
    excludes = _properties.List(_properties.String, 'data.instance.excludes')
    powers = _properties.List(_properties.Integer, 'data.instance.powers')

    typ = _properties.String('data.instance.type', default='ChemicalFormulaFeaturizer',
                             deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 input_descriptor: ChemicalFormulaDescriptor,
                 features: Optional[List[str]] = None,
                 excludes: Optional[List[str]] = None,
                 powers: Optional[List[int]] = None):
        self.name = name
        self.description = description
        self.input_descriptor = input_descriptor
        self.features = features if features is not None else ["standard"]
        self.excludes = excludes if excludes is not None else []
        self.powers = powers if powers is not None else [1]

    def __str__(self):
        return '<ChemicalFormulaFeaturizer {!r}>'.format(self.name)



from typing import List, Optional

from citrine._rest.engine_resource import VersionedEngineResource
from citrine._serialization import properties as _properties
from citrine.informatics.data_sources import DataSource
from citrine.informatics.descriptors import FormulationDescriptor
from citrine.informatics.predictors import Predictor

__all__ = ['SimpleMixturePredictor']


class SimpleMixturePredictor(
        VersionedEngineResource['SimpleMixturePredictor'], Predictor):
    """A predictor interface that flattens a formulation into a simple mixture.

    Parameters
    ----------
    name: str
        name of the configuration
    description: str
        description of the predictor
    input_descriptor: FormulationDescriptor
        input descriptor for the hierarchical (un-mixed) formulation
    output_descriptor: FormulationDescriptor
        output descriptor for the flat (mixed) formulation
    training_data: Optional[List[DataSource]]
        Sources of training data. Each can be either a CSV or an GEM Table. Candidates from
        multiple data sources will be combined into a flattened list and de-duplicated by uid and
        identifiers. De-duplication is performed if a uid or identifier is shared between two or
        more rows. The content of a de-duplicated row will contain the union of data across all
        rows that share the same uid or at least 1 identifier. Training data is unnecessary if the
        predictor is part of a graph that includes all training data required by this predictor.

    """

    input_descriptor = _properties.Object(FormulationDescriptor, 'data.instance.input')
    output_descriptor = _properties.Object(FormulationDescriptor, 'data.instance.output')
    training_data = _properties.List(_properties.Object(DataSource),
                                     'data.instance.training_data', default=[])

    typ = _properties.String('data.instance.type', default='SimpleMixture',
                             deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 input_descriptor: FormulationDescriptor,
                 output_descriptor: FormulationDescriptor,
                 training_data: Optional[List[DataSource]] = None):
        self.name: str = name
        self.description: str = description
        self.input_descriptor: FormulationDescriptor = input_descriptor
        self.output_descriptor: FormulationDescriptor = output_descriptor
        self.training_data: List[DataSource] = training_data or []

    def __str__(self):
        return '<SimpleMixturePredictor {!r}>'.format(self.name)



"""Resources that represent both individual and collections of design workflow executions."""
from typing import Optional, Union, Iterator
from uuid import UUID

from citrine._rest.collection import Collection
from citrine._session import Session
from citrine.informatics.executions.generative_design_execution import GenerativeDesignExecution
from citrine.informatics.generative_design import GenerativeDesignInput
from citrine.resources.response import Response


class GenerativeDesignExecutionCollection(Collection["GenerativeDesignExecution"]):
    """A collection of GenerativeDesignExecutions."""

    _path_template = '/projects/{project_id}/generative-design/executions'
    _individual_key = None
    _collection_key = 'response'
    _resource = GenerativeDesignExecution

    def __init__(self, project_id: UUID, session: Session):
        self.project_id: UUID = project_id
        self.session: Session = session

    def build(self, data: dict) -> GenerativeDesignExecution:
        """Build an individual GenerativeDesignExecution."""
        execution = GenerativeDesignExecution.build(data)
        execution._session = self.session
        execution.project_id = self.project_id
        return execution

    def trigger(
        self, generative_design_execution_input: GenerativeDesignInput
    ) -> GenerativeDesignExecution:
        """Trigger a Generative Design execution."""
        path = self._get_path()
        request_dict = generative_design_execution_input.dump()
        data = self.session.post_resource(path, request_dict)
        return self.build(data)

    def register(self, model: GenerativeDesignExecution) -> GenerativeDesignExecution:
        """Cannot register an execution."""
        raise NotImplementedError("Cannot register a GenerativeDesignExecution.")

    def update(self, model: GenerativeDesignExecution) -> GenerativeDesignExecution:
        """Cannot update an execution."""
        raise NotImplementedError("Cannot update a GenerativeDesignExecution.")

    def list(self, *,
             page: Optional[int] = None,
             per_page: int = 10,
             ) -> Iterator[GenerativeDesignExecution]:
        """
        Paginate over the elements of the collection.

        Leaving page and per_page as default values will yield all elements in the
        collection, paginating over all available pages.

        Parameters
        ---------
        page: int, optional
            The "page" of results to list. Default is to read all pages and yield
            all results.  This option is deprecated.
        per_page: int, optional
            Max number of results to return per page. Default is 100.  This parameter
            is used when making requests to the backend service.  If the page parameter
            is specified it limits the maximum number of elements in the response.

        Returns
        -------
        Iterator[ResourceType]
            Resources in this collection.

        """
        return self._paginator.paginate(page_fetcher=self._fetch_page,
                                        collection_builder=self._build_collection_elements,
                                        page=page,
                                        per_page=per_page)

    def delete(self, uid: Union[UUID, str]) -> Response:
        """Generative Design Executions cannot be deleted or archived."""
        raise NotImplementedError(
            "Generative Design Executions cannot be deleted"
        )



from typing import Generic, TypeVar


Self = TypeVar('Self', bound='Serializable')


class Serializable(Generic[Self]):
    """A Serializable object."""

    @classmethod
    def _pre_build(cls, data: dict) -> dict:
        """Run data modification before building."""
        return data

    @classmethod
    def build(cls, data: dict) -> Self:
        """Build an instance of this object from given data."""
        from citrine._serialization import properties
        pre_built = cls._pre_build(data)
        return properties.Object(cls).deserialize(pre_built)

    def dump(self) -> dict:
        """Dump this instance."""
        from citrine._serialization import properties
        serialized = properties.Object(type(self)).serialize(self)
        return self._post_dump(serialized)

    def _post_dump(self, data: dict) -> dict:
        """Run data modification after dumping."""
        return data



import warnings
from copy import deepcopy
from logging import getLogger
from typing import TypeVar, Optional, Callable

from citrine.exceptions import NotFound
from citrine.resources.team import TeamCollection, Team
from citrine.resources.project import ProjectCollection, Project
from citrine.resources.dataset import DatasetCollection, Dataset
from citrine.informatics.workflows.design_workflow import DesignWorkflow
from citrine._rest.collection import CreationType, Collection

logger = getLogger(__name__)
T = TypeVar('T')


def find_collection(*, collection: Collection[T], name: str) -> Optional[T]:
    """
    Looks through the pages of a collection for a resource with the specified name.

    Returns it, or if not found, returns None
    """
    if isinstance(collection, ProjectCollection):
        try:
            # try to use search if it is available
            # call list() to collapse the iterator, otherwise the NotFound
            # won't show up until collection_list is used
            collection_list = list(collection.search(search_params={
                "name": {
                    "value": name,
                    "search_method": "EXACT"
                }
            }))
        except (NotFound, NotImplementedError):
            # Search must not be available yet or any more
            collection_list = collection.list()
    else:
        collection_list = collection.list()

    matching_resources = [resource for resource in collection_list if resource.name == name]
    if len(matching_resources) > 1:
        raise ValueError("Found multiple collections with name '{}'".format(name))
    if len(matching_resources) == 1:
        result = matching_resources.pop()
        logger.info('Found existing: {}'.format(result))
        return result
    else:
        return None


def get_by_name_or_create(*,
                          collection: Collection[T],
                          name: str,
                          default_provider: Callable[..., T]) -> T:
    """
    Tries to find a collection by its name (returns first hit).

    If not found, implements default_provider
    """
    found = find_collection(collection=collection, name=name)
    if found:
        return found
    else:
        logger.info('Failed to find resource with name {}, creating one instead.'.format(name))
        return default_provider()


def get_by_name_or_raise_error(*, collection: Collection[T], name: str) -> T:
    """
    Tries to find a collection by its name (returns first hit).

    If not found, raises error
    """
    found = find_collection(collection=collection, name=name)
    if found:
        return found
    else:
        raise ValueError("Did not find resource with the given name: {}".format(name))


def find_or_create_project(*,
                           project_collection: ProjectCollection,
                           project_name: str,
                           raise_error: bool = False) -> Project:
    """
    Tries to find a project by name (returns first hit).

    If not found, creates a new project with the given name
    """
    if project_collection.team_id is None:
        # @deprecated(deprecated_in="1.33.1", removed_in="2.0.0", ...)
        warnings.warn("This method will be unreliable once Teams are released, "
                      "at which point you should use find_or_create_team.",
                      DeprecationWarning)
    if raise_error:
        project = get_by_name_or_raise_error(collection=project_collection, name=project_name)
    else:
        project = get_by_name_or_create(
            collection=project_collection,
            name=project_name,
            default_provider=lambda: project_collection.register(project_name)
        )
    return project


def find_or_create_team(*,
                        team_collection: TeamCollection,
                        team_name: str,
                        raise_error: bool = False) -> Team:
    """
    Tries to find a team by name (returns first hit).

    If not found, creates a new team with the given name
    """
    if raise_error:
        team = get_by_name_or_raise_error(collection=team_collection, name=team_name)
    else:
        team = get_by_name_or_create(
            collection=team_collection,
            name=team_name,
            default_provider=lambda: team_collection.register(team_name)
        )
    return team


def find_or_create_dataset(*,
                           dataset_collection: DatasetCollection,
                           dataset_name: str,
                           raise_error: bool = False) -> Dataset:
    """
    Tries to find a dataset by name (returns first hit).

    If not found, creates a new dataset with the given name
    """
    if raise_error:
        dataset = get_by_name_or_raise_error(collection=dataset_collection, name=dataset_name)
    else:
        dataset = get_by_name_or_create(
            collection=dataset_collection,
            name=dataset_name,
            default_provider=lambda: dataset_collection.register(
                Dataset(dataset_name, summary="seed summ.", description="seed desc.")
            )
        )
    return dataset


def create_or_update(*,
                     collection: Collection[CreationType],
                     resource: CreationType) -> CreationType:
    """
    Update a resource of a given name belonging to a collection.

    Create it if it doesn't exist. If collection already contains
    a resource with the same name it will be updated. If there are multiple
    resources with the same name it will throw an error.

    Parameters
    ----------
    collection : Collection[CreationType]
        Collection within which you want to update or create a resource
    resource : CreationType
        Resource that you want to create or update.

    Returns
    -------
    CreationType
        Registered updated or created resource.

    """
    old_resource = find_collection(collection=collection, name=resource.name)
    if old_resource:
        logger.info("Updating module: {}".format(resource.name))
        # Copy so that passed-in resource is unaffected
        new_resource = deepcopy(resource)
        new_resource.uid = old_resource.uid
        # Locally created design workflows likely won't have a branch ID but
        # need one to be updated.
        if isinstance(old_resource, DesignWorkflow):
            new_resource.branch_id = old_resource.branch_id
        return collection.update(new_resource)
    else:
        logger.info("Registering new module:  {}".format(resource.name))
        return collection.register(resource)



"""Tools for working with Scores."""
import warnings
from typing import List, Optional

from citrine._serialization import properties
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization.serializable import Serializable
from citrine.informatics.constraints import Constraint
from citrine.informatics.objectives import Objective

__all__ = ['Score', 'LIScore', 'EIScore', 'EVScore']


class Score(PolymorphicSerializable['Score']):
    """A Score is used to rank materials according to objectives and constraints.

    Abstract type that returns the proper type given a serialized dict.

    """

    _name = properties.String('name')
    _description = properties.String('description')

    @classmethod
    def get_type(cls, data):
        """Return the subtype."""
        return {
            'MLI': LIScore,
            'MEI': EIScore,
            'MEV': EVScore
        }[data['type']]

    @property
    def name(self):
        """Getter for the score's name."""
        msg = "Getting the Score's name is deprecated."
        warnings.warn(msg, category=DeprecationWarning)
        return self._name

    @property
    def description(self):
        """Getter for the score's description."""
        msg = "Getting the Score's description is deprecated."
        warnings.warn(msg, category=DeprecationWarning)
        return self._description


class LIScore(Serializable['LIScore'], Score):
    """Evaluates the likelihood of scoring better than some baselines for given objectives.

    Parameters
    ----------
    objectives: list[Objective]
        objectives (e.g., maximize, minimize, tune, etc.)
        If multiple objectives are specified then this score evaluates the likelihood of
        simultaneously exceeding all objectives.
    baselines: list[float]
        best-so-far values for the various objectives (there must be one for each objective)
    constraints: list[Constraint]
        constraints limiting the allowed values that material instances can have

    """

    baselines = properties.List(properties.Float, 'baselines')
    objectives = properties.List(properties.Object(Objective), 'objectives')
    constraints = properties.List(properties.Object(Constraint), 'constraints')
    typ = properties.String('type', default='MLI')

    def __init__(self, *,
                 name: Optional[str] = None,
                 description: Optional[str] = None,
                 objectives: List[Objective],
                 baselines: List[float],
                 constraints: Optional[List[Constraint]] = None):
        self.objectives: List[Objective] = objectives
        self.baselines: List[float] = baselines
        self.constraints: List[Constraint] = constraints or []

        if name is not None:
            msg = "Naming of Scores is deprecated.  Please do not define the name."
            warnings.warn(msg, category=DeprecationWarning)
            self._name = name
        else:
            self._name = "Likelihood of Improvement"

        if description is not None:
            msg = "Describing Scores is deprecated.  Please do not define the description."
            warnings.warn(msg, category=DeprecationWarning)
            self._description: str = description
        else:
            self._description = ""

    def __str__(self):
        return '<LIScore>'


class EIScore(Serializable['EIScore'], Score):
    """
    Evaluates the expected magnitude of improvement beyond baselines for a given objective.

    Parameters
    ----------
    objectives: list[Objective]
        objectives (e.g., maximize, minimize, tune, etc.)
        EIScore does not support more than 1 objective at this time.
    baselines: list[float]
        best-so-far values for the various objectives (there must be one for each objective)
    constraints: list[Constraint]
        constraints limiting the allowed values that material instances can have

    """

    baselines = properties.List(properties.Float, 'baselines')
    objectives = properties.List(properties.Object(Objective), 'objectives')
    constraints = properties.List(properties.Object(Constraint), 'constraints')
    typ = properties.String('type', default='MEI')

    def __init__(self, *,
                 name: Optional[str] = None,
                 description: Optional[str] = None,
                 objectives: List[Objective],
                 baselines: List[float],
                 constraints: Optional[List[Constraint]] = None):
        self.objectives: List[Objective] = objectives
        self.baselines: List[float] = baselines
        self.constraints: List[Constraint] = constraints or []

        if name is not None:
            msg = "Naming of Scores is deprecated.  Please do not define the name."
            warnings.warn(msg, category=DeprecationWarning)
            self._name = name
        else:
            self._name = "Expected Improvement"

        if description is not None:
            msg = "Describing Scores is deprecated.  Please do not define the description."
            warnings.warn(msg, category=DeprecationWarning)
            self._description: str = description
        else:
            self._description = ""

    def __str__(self):
        return '<EIScore>'


class EVScore(Serializable['EVScore'], Score):
    """
    Evaluates the expected value for given objectives.

    Parameters
    ----------
    objectives: list[Objective]
        objectives (e.g., maximize, minimize, tune, etc.)
        If multiple objectives are specified, their scores are summed together. This allows
        for simultaneous optimization of multiple objectives, although the weighting of the
        various objectives cannot be directly specified.
    constraints: list[Constraint]
        constraints limiting the allowed values that material instances can have

    """

    objectives = properties.List(properties.Object(Objective), 'objectives')
    constraints = properties.List(properties.Object(Constraint), 'constraints')
    typ = properties.String('type', default='MEV')

    def __init__(self, *,
                 name: Optional[str] = None,
                 description: Optional[str] = None,
                 objectives: List[Objective],
                 constraints: Optional[List[Constraint]] = None):
        self.objectives: List[Objective] = objectives
        self.constraints: List[Constraint] = constraints or []

        if name is not None:
            msg = "Naming of Scores is deprecated.  Please do not define the name."
            warnings.warn(msg, category=DeprecationWarning)
            self._name = name
        else:
            self._name = "Expected Value"

        if description is not None:
            msg = "Describing Scores is deprecated.  Please do not define the description."
            warnings.warn(msg, category=DeprecationWarning)
            self._description: str = description
        else:
            self._description = ""

    def __str__(self):
        return '<EVScore>'



"""A resource that represents a single module report."""
from typing import Optional, Union
from uuid import UUID

from deprecation import deprecated

from citrine._rest.resource import Resource
from citrine._session import Session
from citrine._utils.functions import format_escaped_url
from citrine._utils.functions import migrate_deprecated_argument
from citrine.informatics.reports import Report


class ReportResource(Resource['ReportResource']):
    """Defines a resource for fetching reports from a module.

    Parameters
    ----------
    project_id: UUID
        the UUID of the project

    """

    _path_template = '/projects/{project_id}/predictors/{predictor_id}/versions/{version}/report'
    _api_version = 'v3'

    def __init__(self, project_id: UUID, session: Session):
        self.project_id = project_id
        self.session = session

    def get(self,
            module_id: UUID = None,
            *,
            predictor_id: Union[UUID, str] = None,
            predictor_version: Optional[Union[int, str]] = None) -> Report:
        """Gets a single report keyed on the predictor ID and (optionally) version."""
        predictor_id = migrate_deprecated_argument(predictor_id, "predictor_id",
                                                   module_id, "module_id")
        version = predictor_version or "most_recent"

        url_path = format_escaped_url(self._path_template,
                                      project_id=self.project_id,
                                      predictor_id=str(predictor_id),
                                      version=version)
        data = self.session.get_resource(url_path, version=self._api_version)

        report = Report.build(data)
        report.session = self.session
        return report

    @deprecated(deprecated_in="1.49.0", removed_in="2.0.0",
                details="Please use ReportResource.get() instead.")
    def get_version(self, predictor_id: Union[UUID, str], *, predictor_version: Union[int, str]) \
            -> Report:
        """Gets a single report keyed on the predictor ID and version."""
        return self.get(predictor_id=predictor_id, predictor_version=predictor_version)



from typing import Mapping

from citrine._rest.engine_resource import VersionedEngineResource
from citrine._serialization import properties as _properties
from citrine.informatics.descriptors import RealDescriptor
from citrine.informatics.predictors import Predictor

__all__ = ['ExpressionPredictor']


class ExpressionPredictor(VersionedEngineResource['ExpressionPredictor'], Predictor):
    """A predictor that computes an output from an expression and set of bounded inputs.

    For a discussion of expression syntax and a list of allowed symbols,
    please see the :ref:`documentation<Expression Predictor>`.

    Parameters
    ----------
    name: str
        name of the configuration
    description: str
        the description of the predictor
    expression: str
        expression that computes an output from aliased inputs
    output: RealDescriptor
        descriptor that represents the output of the expression
    aliases: Mapping[str, RealDescriptor]
        a mapping from each unknown argument to its descriptor.
        All unknown arguments must have an associated descriptor.

    """

    expression = _properties.String('data.instance.expression')
    output = _properties.Object(RealDescriptor, 'data.instance.output')
    aliases = _properties.Mapping(_properties.String, _properties.Object(RealDescriptor),
                                  'data.instance.aliases')

    typ = _properties.String('data.instance.type',
                             default='AnalyticExpression', deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 expression: str,
                 output: RealDescriptor,
                 aliases: Mapping[str, RealDescriptor]):
        self.name: str = name
        self.description: str = description
        self.expression: str = expression
        self.output: RealDescriptor = output
        self.aliases: Mapping[str, RealDescriptor] = aliases

    def __str__(self):
        return '<ExpressionPredictor {!r}>'.format(self.name)



from deprecation import deprecated
from typing import List, TypeVar

from citrine._rest.resource import Resource, ResourceTypeEnum
from citrine._serialization import properties
from citrine._serialization.include_parent_properties import IncludeParentProperties
from citrine.resources.status_detail import StatusDetail

Self = TypeVar('Self', bound='Resource')


# This class is the base type for the new module endpoints which do not support versions. If/once
# they support versioning, they should be switched to inherit from VersionedEngineResource.
class EngineResource(Resource[Self]):
    """Base resource for metadata from stand-alone AI Engine modules."""

    created_by = properties.Optional(properties.UUID, 'metadata.created.user', serializable=False)
    """:Optional[UUID]: id of the user who created the resource"""
    create_time = properties.Optional(properties.Datetime, 'metadata.created.time',
                                      serializable=False)
    """:Optional[datetime]: date and time at which the resource was created"""

    updated_by = properties.Optional(properties.UUID, 'metadata.updated.user',
                                     serializable=False)
    """:Optional[UUID]: id of the user who most recently updated the resource,
    if it has been updated"""
    update_time = properties.Optional(properties.Datetime, 'metadata.updated.time',
                                      serializable=False)
    """:Optional[datetime]: date and time at which the resource was most recently updated,
    if it has been updated"""

    archived_by = properties.Optional(properties.UUID, 'metadata.archived.user',
                                      serializable=False)
    """:Optional[UUID]: id of the user who archived the resource, if it has been archived"""
    archive_time = properties.Optional(properties.Datetime, 'metadata.archived.time',
                                       serializable=False)
    """:Optional[datetime]: date and time at which the resource was archived,
    if it has been archived"""

    status = properties.Optional(properties.String(), 'metadata.status.name', serializable=False)
    """:Optional[str]: short description of the resource's status"""
    status_detail = properties.List(properties.Object(StatusDetail), 'metadata.status.detail',
                                    default=[], serializable=False)
    """:List[StatusDetail]: a list of structured status info, containing the message and level"""

    # Due to the way object construction is done at present, __init__ is not executed on Resource
    # objects, so initializing _archived doesn't work.
    _archived = properties.Optional(properties.Boolean(), '', default=None, serializable=False,
                                    deserializable=False)

    _resource_type = ResourceTypeEnum.MODULE

    @property
    def is_archived(self):
        """:bool: whether the resource is archived (hidden but not deleted)."""
        return self.archived_by is not None

    @property
    @deprecated(deprecated_in="1.46.0", removed_in="2.0.0",
                details="Please use the 'status_detail' property instead.'")
    def status_info(self):
        """[DEPRECATED] :list[str]: human-readable explanations of the status."""
        return [detail.msg for detail in self.status_detail]

    @property
    @deprecated(deprecated_in="1.31.0", removed_in="2.0.0",
                details="Please use the 'is_archived' property instead.'")
    def archived(self):
        """[DEPRECATED] whether the resource is archived."""
        return self.is_archived

    @archived.setter
    @deprecated(deprecated_in="1.31.0", removed_in="2.0.0",
                details="Please use archive() and restore() on PredictorCollection instead.")
    def archived(self, value):
        self._archived = value

    @property
    @deprecated(deprecated_in="1.25.0", removed_in="2.0.0")
    def experimental(self) -> bool:
        """[DEPRECATED] whether the execution is experimental (newer, less well-tested)."""  # noqa - insisting this docstring is a signature
        return False

    @property
    @deprecated(deprecated_in="1.25.0", removed_in="2.0.0")
    def experimental_reasons(self) -> List[str]:
        """[DEPRECATED] human-readable reasons why the execution is experimental."""
        return []

    def _post_dump(self, data: dict) -> dict:
        # Only the data portion of an entity is sent to the server.
        data = data["data"]

        # Currently, name and description exists on both the data envelope and the config.
        data["instance"]["name"] = data["name"]
        data["instance"]["description"] = data["description"]
        return data


class VersionedEngineResource(EngineResource[Self], IncludeParentProperties[Self]):
    """Base resource for metadata from stand-alone AI Engine modules which support versioning."""

    """:Integer: The version number of the resource."""
    version = properties.Optional(properties.Integer, 'metadata.version', serializable=False)

    """:Boolean: The draft status of the resource."""
    draft = properties.Optional(properties.Boolean, 'metadata.draft', serializable=False)

    @classmethod
    def build(cls, data: dict):
        """Build an instance of this object from given data."""
        return super().build_with_parent(data, __class__)



"""Resources that represent both individual and collections of workflow executions."""
from typing import Optional, Union
from uuid import UUID

from citrine._rest.collection import Collection
from citrine._session import Session
from citrine._utils.functions import migrate_deprecated_argument
from citrine.informatics.workflows import PredictorEvaluationWorkflow
from citrine.resources.response import Response


class PredictorEvaluationWorkflowCollection(Collection[PredictorEvaluationWorkflow]):
    """A collection of PredictorEvaluationWorkflows."""

    _path_template = '/projects/{project_id}/predictor-evaluation-workflows'
    _individual_key = None
    _collection_key = 'response'
    _resource = PredictorEvaluationWorkflow

    def __init__(self, project_id: UUID, session: Session):
        self.project_id: UUID = project_id
        self.session: Session = session

    def build(self, data: dict) -> PredictorEvaluationWorkflow:
        """Build an individual PredictorEvaluationExecution."""
        workflow = PredictorEvaluationWorkflow.build(data)
        workflow._session = self.session
        workflow.project_id = self.project_id
        return workflow

    def archive(self, uid: Union[UUID, str] = None, workflow_id: Union[UUID, str] = None):
        """Archive a predictor evaluation workflow.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the workflow to archive
        workflow_id: Union[UUID, str]
            [DEPRECATED] please use uid instead

        """
        uid = migrate_deprecated_argument(uid, "uid", workflow_id, "workflow_id")
        return self._put_resource_ref('archive', uid)

    def restore(self, uid: Union[UUID, str] = None, workflow_id: Union[UUID, str] = None):
        """Restore an archived predictor evaluation workflow.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the workflow to restore
        workflow_id: Union[UUID, str]
            [DEPRECATED] please use uid instead

        """
        uid = migrate_deprecated_argument(uid, "uid", workflow_id, "workflow_id")
        return self._put_resource_ref('restore', uid)

    def delete(self, uid: Union[UUID, str]) -> Response:
        """Predictor Evaluation Workflows cannot be deleted; they can be archived instead."""
        raise NotImplementedError(
            "Predictor Evaluation Workflows cannot be deleted; they can be archived instead.")

    def create_default(self,
                       *,
                       predictor_id: UUID,
                       predictor_version: Optional[Union[int, str]] = None) \
            -> PredictorEvaluationWorkflow:
        """[ALPHA] Create a default predictor evaluation workflow for a predictor and execute it.

        The current default predictor evaluation workflow performs 5-fold, 1-trial cross-validation
        on all valid predictor responses. Valid responses are those that are **not** produced by the
        following predictors:

        * :class:`~citrine.informatics.predictors.generalized_mean_property_predictor.GeneralizedMeanPropertyPredictor`
        * :class:`~citrine.informatics.predictors.mean_property_predictor.MeanPropertyPredictor`
        * :class:`~citrine.informatics.predictors.ingredient_fractions_predictor.IngredientFractionsPredictor`
        * :class:`~citrine.informatics.predictors.ingredients_to_simple_mixture_predictor.IngredientsToSimpleMixturePredictor`
        * :class:`~citrine.informatics.predictors.ingredients_to_formulation_predictor.IngredientsToFormulationPredictor`
        * :class:`~citrine.informatics.predictors.label_fractions_predictor.LabelFractionsPredictor`
        * :class:`~citrine.informatics.predictors.molecular_structure_featurizer.MolecularStructureFeaturizer`
        * :class:`~citrine.informatics.predictors.simple_mixture_predictor.SimpleMixturePredictor`

        If there are no valid responses, a default workflow is not created.

        Parameters
        ----------
        predictor_id: UUID
            Unique identifier of the predictor used to create a default workflow
        predictor_version: Option[Union[int, str]]
            The version of the predictor used to create a default workflow

        Returns
        -------
        PredictorEvaluationWorkflow
            Default workflow

        """  # noqa: E501,W505
        url = self._get_path('default')
        payload = {'predictor_id': str(predictor_id)}
        if predictor_version:
            payload['predictor_version'] = predictor_version
        data = self.session.post_resource(url, payload)
        return self.build(data)



"""Resources that represent both individual and collections of predictor evaluation executions."""
from functools import partial
import sys
from typing import Optional, Union, Iterator
from uuid import UUID

from citrine._rest.collection import Collection
from citrine._rest.resource import PredictorRef
from citrine._session import Session
from citrine._utils.functions import migrate_deprecated_argument, shadow_classes_in_module, \
    format_escaped_url
from citrine.informatics.executions import PredictorEvaluationExecution
import citrine.informatics.executions.predictor_evaluation_execution
from citrine.resources.response import Response


shadow_classes_in_module(citrine.informatics.executions.predictor_evaluation_execution,
                         sys.modules[__name__])


class PredictorEvaluationExecutionCollection(Collection["PredictorEvaluationExecution"]):
    """A collection of PredictorEvaluationExecutions."""

    _path_template = '/projects/{project_id}/predictor-evaluation-executions'  # noqa
    _individual_key = None
    _collection_key = 'response'
    _resource = PredictorEvaluationExecution

    def __init__(self,
                 project_id: UUID,
                 session: Session,
                 workflow_id: Optional[UUID] = None):
        self.project_id: UUID = project_id
        self.session: Session = session
        self.workflow_id: Optional[UUID] = workflow_id

    def build(self, data: dict) -> PredictorEvaluationExecution:
        """Build an individual PredictorEvaluationExecution."""
        execution = PredictorEvaluationExecution.build(data)
        execution._session = self.session
        execution.project_id = self.project_id
        return execution

    def trigger(self,
                predictor_id: UUID,
                *,
                predictor_version: Optional[Union[int, str]] = None,
                random_state: Optional[int] = None):
        """Trigger a predictor evaluation execution against a predictor.

        Parameters
        -----------
        predictor_id: UUID
            ID of the predictor to evaluate.
        predictor_version: Union[int, str], optional
            The version of the predictor to evaluate.
        random_state: int, optional
            Seeds the evaluators' random number generator so that the results are repeatable.

        """
        if self.workflow_id is None:
            msg = "Cannot trigger a predictor evaluation execution without knowing the " \
                  "predictor evaluation workflow. Use workflow.executions.trigger instead of " \
                  "project.predictor_evaluation_executions.trigger"
            raise RuntimeError(msg)
        path = format_escaped_url(
            '/projects/{project_id}/predictor-evaluation-workflows/{workflow_id}/executions',
            project_id=self.project_id,
            workflow_id=self.workflow_id
        )

        params = dict()
        if random_state is not None:
            params["random_state"] = random_state

        payload = PredictorRef(predictor_id, predictor_version).dump()
        data = self.session.post_resource(path, payload, params=params, version='v2')

        return self.build(data)

    def register(self, model: PredictorEvaluationExecution) -> PredictorEvaluationExecution:
        """Cannot register an execution."""
        raise NotImplementedError("Cannot register a PredictorEvaluationExecution.")

    def update(self, model: PredictorEvaluationExecution) -> PredictorEvaluationExecution:
        """Cannot update an execution."""
        raise NotImplementedError("Cannot update a PredictorEvaluationExecution.")

    def archive(self, uid: Union[UUID, str] = None, execution_id: Union[UUID, str] = None):
        """Archive a predictor evaluation execution.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the execution to archive
        execution_id: Union[UUID, str]
            [DEPRECATED] please use uid instead

        """
        uid = migrate_deprecated_argument(uid, "uid", execution_id, "execution_id")
        self._put_resource_ref('archive', uid)

    def restore(self, uid: Union[UUID, str] = None, execution_id: Union[UUID, str] = None):
        """Restore an archived predictor evaluation execution.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the execution to restore
        execution_id: Union[UUID, str]
            [DEPRECATED] please use uid instead

        """
        uid = migrate_deprecated_argument(uid, "uid", execution_id, "execution_id")
        self._put_resource_ref('restore', uid)

    def list(self,
             *,
             page: Optional[int] = None,
             per_page: int = 100,
             predictor_id: Optional[UUID] = None,
             predictor_version: Optional[Union[int, str]] = None
             ) -> Iterator[PredictorEvaluationExecution]:
        """
        Paginate over the elements of the collection.

        Leaving page and per_page as default values will yield all elements in the
        collection, paginating over all available pages.

        Parameters
        ---------
        page: int, optional
            The "page" of results to list. Default is to read all pages and yield
            all results.  This option is deprecated.
        per_page: int, optional
            Max number of results to return per page. Default is 100.  This parameter
            is used when making requests to the backend service.  If the page parameter
            is specified it limits the maximum number of elements in the response.
        predictor_id: uuid, optional
            list executions that targeted the predictor with this id
        predictor_version: Union[int, str], optional
            list executions that targeted the predictor with this version

        Returns
        -------
        Iterator[PredictorEvaluationExecution]
            The matching predictor evaluation executions.

        """
        params = {}
        if predictor_id is not None:
            params["predictor_id"] = str(predictor_id)
        if predictor_version is not None:
            params["predictor_version"] = predictor_version
        if self.workflow_id is not None:
            params["workflow_id"] = str(self.workflow_id)

        fetcher = partial(self._fetch_page, additional_params=params)
        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_collection_elements,
                                        page=page,
                                        per_page=per_page)

    def delete(self, uid: Union[UUID, str]) -> Response:
        """Predictor Evaluation Executions cannot be deleted; they can be archived instead."""
        raise NotImplementedError(
            "Predictor Evaluation Executions cannot be deleted; they can be archived instead.")



"""Tools for working with Processors."""
from typing import Optional, Mapping, Type, List

from citrine._rest.resource import ResourceTypeEnum, Resource
from citrine._rest.ai_resource_metadata import AIResourceMetadata
from citrine._serialization import properties
from citrine.informatics.modules import Module


__all__ = ['Processor', 'GridProcessor', 'EnumeratedProcessor', 'MonteCarloProcessor']


class Processor(Module):
    """A Citrine Processor describes how a design space is searched.

    Abstract type that returns the proper type given a serialized dict.

    """

    uid = properties.Optional(properties.UUID, 'id', serializable=False)
    """:Optional[UUID]: Citrine Platform unique identifier"""
    name = properties.String('config.name')
    description = properties.Optional(properties.String(), 'config.description')

    @classmethod
    def get_type(cls, data) -> Type['Processor']:
        """Return the sole currently implemented subtype."""
        return {
            'Grid': GridProcessor,
            'Enumerated': EnumeratedProcessor,
            'ContinuousSearch': MonteCarloProcessor
        }[data['config']['type']]

    def _attrs(self) -> List[str]:
        return ["name", "description", "id"]  # pragma: no cover

    def __eq__(self, other):
        try:
            return all([
                self.__getattribute__(key) == other.__getattribute__(key) for key in self._attrs()
            ])
        except AttributeError:
            return False


class GridProcessor(Resource['GridProcessor'], Processor, AIResourceMetadata):
    """Generates samples from the Cartesian product of finite dimensions, then scans over them.

    For each continuous dimensions, a uniform grid is created between the lower and upper bounds of
    the descriptor. The number of points along each continuous dimension is specified.
    by ``grid_sizes``. No such discretization is necessary for enumerated dimensions,
    because they are finite.

    Be careful when using a grid processor, as the number of points grows exponentially with
    the number of dimensions. For high-dimensional design spaces, a continuous processor
    is often preferable.

    Parameters
    ----------
    name: str
        name of the processor
    description: str
        description of the processor
    grid_sizes: dict[str, int]
        the number of points to select along each continuous dimension, by dimension name

    """

    _resource_type = ResourceTypeEnum.MODULE

    grid_sizes = properties.Mapping(
        properties.String,
        properties.Integer,
        'config.grid_dimensions'
    )

    typ = properties.String('config.type', default='Grid', deserializable=False)
    module_type = properties.String('module_type', default='PROCESSOR')

    def _attrs(self) -> List[str]:
        return ["name", "description", "grid_sizes", "typ"]

    def __init__(self,
                 name: str, *,
                 description: str,
                 grid_sizes: Mapping[str, int]):
        self.name: str = name
        self.description: str = description
        self.grid_sizes: Mapping[str, int] = grid_sizes

    def _post_dump(self, data: dict) -> dict:
        data['display_name'] = data['config']['name']
        return data

    def __str__(self):
        return '<GridProcessor {!r}>'.format(self.name)


class EnumeratedProcessor(Resource['EnumeratedProcessor'], Processor, AIResourceMetadata):
    """Process a design space by enumerating up to a fixed number of samples from the domain.

    Each sample is processed independently.

    Parameters
    ----------
    name: str
        name of the processor
    description: str
        description of the processor
    max_candidates: int
        maximum number of samples that can be enumerated over (default: 1000)

    """

    _resource_type = ResourceTypeEnum.MODULE

    max_candidates = properties.Integer('config.max_size')

    typ = properties.String('config.type', default='Enumerated', deserializable=False)
    module_type = properties.String('module_type', default='PROCESSOR')

    def _attrs(self) -> List[str]:
        return ["name", "description", "max_candidates", "typ"]

    def __init__(self,
                 name: str, *,
                 description: str,
                 max_candidates: Optional[int] = None):
        self.name: str = name
        self.description: str = description
        self.max_candidates: int = max_candidates or 1000

    def _post_dump(self, data: dict) -> dict:
        data['display_name'] = data['config']['name']
        return data

    def __str__(self):
        return '<EnumeratedProcessor {!r}>'.format(self.name)


class MonteCarloProcessor(Resource['ContinuousSearch'], Processor, AIResourceMetadata):
    """Using a Monte Carlo optimizer to search for the best candidate.

    The moves that the MonteCarlo optimizer makes are inferred from the descriptors in the
    design space.

    Parameters
    ----------
    name: str
        name of the processor
    description: str
        description of the processor
    max_candidates: int
        maximum number of candidates generated by the processor (default: system configured limit)
    mode: str
        for internal use only

    """

    _resource_type = ResourceTypeEnum.MODULE

    max_candidates = properties.Optional(properties.Integer, 'config.max_candidates')
    mode = properties.Optional(properties.String(), 'config.mode')

    typ = properties.String('config.type', default='ContinuousSearch', deserializable=False)
    module_type = properties.String('module_type', default='PROCESSOR')

    def _attrs(self) -> List[str]:
        return ["name", "description", "mode", "typ"]

    def __init__(self,
                 name: str, *,
                 description: str,
                 max_candidates: Optional[int] = None,
                 mode: Optional[str] = None):
        self.name: str = name
        self.description: str = description
        self.max_candidates: Optional[int] = max_candidates
        self.mode: Optional[str] = mode

    def _post_dump(self, data: dict) -> dict:
        data['display_name'] = data['config']['name']
        return data

    def __str__(self):
        return '<MonteCarloProcessor {!r}>'.format(self.name)



import sys

import citrine.jobs.job
from citrine.jobs.job import _poll_for_job_completion  # noqa
from citrine._utils.functions import shadow_classes_in_module

shadow_classes_in_module(citrine.jobs.job, sys.modules[__name__])



from typing import Dict, Type

from citrine._serialization.serializable import Serializable
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization import properties

__all__ = [
    'ExperimentValue',
    'RealExperimentValue',
    'IntegerExperimentValue',
    'CategoricalExperimentValue',
    'MixtureExperimentValue',
    'ChemicalFormulaExperimentValue',
    'MolecularStructureExperimentValue'
]


class ExperimentValue(PolymorphicSerializable['ExperimentValue']):
    """An container for experiment values.

    Abstract type that returns the proper type given a serialized dict.
    """

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        return {
            "RealValue": RealExperimentValue,
            "IntegerValue": IntegerExperimentValue,
            "CategoricalValue": CategoricalExperimentValue,
            "MixtureValue": MixtureExperimentValue,
            "OrganicValue": ChemicalFormulaExperimentValue,
            "InorganicValue": MolecularStructureExperimentValue,
        }[data["type"]]

    def __str__(self):
        return f"<{self.__class__.__name__} {self.value!r}>"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.value})"

    def __eq__(self, other):
        return self._equals(other, ["value", "typ"])

    def _equals(self, other, attrs) -> bool:
        """Check to see if the attrs from the other instance match this instance.

        Returns True if all of the attribute names from attrs match between this and
        the other instance, else False. A missing attribute from this instance will
        raise an AttributeError, and False if missing from the other instance.

        Parameters
        ----------
        other: Description
            the Description instance to compare to
        attrs: List[str]
            A list of attribute names to lookup and compare

        """
        # Check that all attrs exist on this object and raise an AttributeError if not.
        [self.__getattribute__(key) for key in attrs]

        try:
            return all([
                self.__getattribute__(key) == other.__getattribute__(key) for key in attrs
            ])
        except AttributeError:
            return False


class RealExperimentValue(Serializable['RealExperimentValue'], ExperimentValue):
    """A floating point experiment result."""

    value = properties.Float('value')
    typ = properties.String('type', default='RealValue', deserializable=False)

    def __init__(self, value: float):
        self.value = value


class IntegerExperimentValue(Serializable['IntegerExperimentValue'], ExperimentValue):
    """An integer value experiment result."""

    value = properties.Integer('value')
    typ = properties.String('type', default='IntegerValue', deserializable=False)

    def __init__(self, value: int):
        self.value = value


class CategoricalExperimentValue(Serializable['CategoricalExperimentValue'], ExperimentValue):
    """An experiment result with a categorical value."""

    value = properties.String('value')
    typ = properties.String('type', default='CategoricalValue', deserializable=False)

    def __init__(self, value: str):
        self.value = value


class MixtureExperimentValue(Serializable['MixtureExperimentValue'], ExperimentValue):
    """An experiment result mapping ingredients and labels to real values."""

    value = properties.Mapping(properties.String, properties.Float, 'value')
    typ = properties.String('type', default='MixtureValue', deserializable=False)

    def __init__(self, value: Dict[str, float]):
        self.value = value


class ChemicalFormulaExperimentValue(Serializable['ChemicalFormulaExperimentValue'],
                                     ExperimentValue):
    """Experiment value for a chemical formula."""

    value = properties.String('value')
    typ = properties.String('type', default='OrganicValue', deserializable=False)

    def __init__(self, value: str):
        self.value = value


class MolecularStructureExperimentValue(Serializable['MolecularStructureExperimentValue'],
                                        ExperimentValue):
    """Experiment value for a molecular structure."""

    value = properties.String('value')
    typ = properties.String('type', default='InorganicValue', deserializable=False)

    def __init__(self, value: str):
        self.value = value



"""Resources that represent collections of predictors."""
from functools import partial
from typing import Any, Iterable, Optional, TypeVar, Union
from uuid import UUID

from gemd.enumeration.base_enumeration import BaseEnumeration

from citrine._rest.paginator import Paginator
from citrine._session import Session
from citrine._utils.functions import migrate_deprecated_argument, format_escaped_url
from citrine.exceptions import Conflict
from citrine.informatics.data_sources import DataSource
from citrine.informatics.predictors import Predictor
from citrine.resources.module import AbstractModuleCollection

from deprecation import deprecated

CreationType = TypeVar('CreationType', bound=Predictor)


MOST_RECENT_VER = "most_recent"


class AutoConfigureMode(BaseEnumeration):
    """[ALPHA] The format to use in building auto-configured assets.

    * PLAIN corresponds to a single-row GEM table and plain predictor
    * FORMULATION corresponds to a multi-row GEM table and formulations predictor
    * INFER auto-detects the GEM table and predictor type
    """

    PLAIN = 'PLAIN'
    FORMULATION = 'FORMULATION'
    INFER = 'INFER'


class _PredictorVersionPaginator(Paginator):
    def _comparison_fields(self, entity: Predictor) -> Any:
        return (entity.uid, entity.version)

    def paginate(self, *args, **kwargs) -> Iterable[Predictor]:
        # Since predictor versions have the same uid, and the paginate method uses uid alone to
        # dedup, we have to disable deduplication in order to use it.
        kwargs["deduplicate"] = False
        return super().paginate(*args, **kwargs)


class _PredictorVersionCollection(AbstractModuleCollection[Predictor]):
    _api_version = 'v3'
    _path_template = '/projects/{project_id}/predictors/{uid}/versions'
    _individual_key = None
    _resource = Predictor
    _collection_key = 'response'
    _paginator: Paginator = _PredictorVersionPaginator()

    _SPECIAL_VERSIONS = ["latest", MOST_RECENT_VER]

    def __init__(self, project_id: UUID, session: Session):
        self.project_id = project_id
        self.session: Session = session

    def _construct_path(self,
                        uid: Union[UUID, str],
                        version: Optional[Union[int, str]] = None,
                        action: str = None) -> str:
        path = self._path_template.format(project_id=self.project_id, uid=str(uid))
        if version is not None:
            version_str = str(version)
            if version_str not in self._SPECIAL_VERSIONS \
                    and (not version_str.isdecimal() or int(version_str) <= 0):
                raise ValueError("A predictor version must either be a positive integer, "
                                 f"\"latest\", or \"{MOST_RECENT_VER}\".")

            path += f"/{version_str}"
            path += f"/{action}" if action else ""
        return path

    def _page_fetcher(self, *, uid: Union[UUID, str], **additional_params):
        fetcher_params = {
            "path": self._construct_path(uid),
            "additional_params": additional_params
        }
        return partial(self._fetch_page, **fetcher_params)

    def _train(self, uid: Union[UUID, str], version: Union[int, str]) -> Predictor:
        path = self._construct_path(uid, version, "train")
        params = {"create_version": True}
        entity = self.session.put_resource(path, {}, params=params, version=self._api_version)
        return self.build(entity)

    def build(self, data: dict) -> Predictor:
        """Build an individual Predictor."""
        predictor: Predictor = Predictor.build(data)
        predictor._session = self.session
        predictor._project_id = self.project_id
        return predictor

    def get(self,
            uid: Union[UUID, str],
            *,
            version: Union[int, str] = MOST_RECENT_VER) -> Predictor:
        path = self._construct_path(uid, version)
        entity = self.session.get_resource(path, version=self._api_version)
        return self.build(entity)

    def list(self,
             uid: Union[UUID, str],
             *,
             page: Optional[int] = None,
             per_page: int = 100) -> Iterable[Predictor]:
        """List non-archived versions of the given predictor."""
        page_fetcher = self._page_fetcher(uid=uid)
        return self._paginator.paginate(page_fetcher=page_fetcher,
                                        collection_builder=self._build_collection_elements,
                                        page=page,
                                        per_page=per_page)

    def list_archived(self,
                      uid: Union[UUID, str],
                      *,
                      page: Optional[int] = None,
                      per_page: int = 20) -> Iterable[Predictor]:
        """List archived versions of the given predictor."""
        page_fetcher = self._page_fetcher(uid=uid, filter="archived eq 'true'")
        return self._paginator.paginate(page_fetcher=page_fetcher,
                                        collection_builder=self._build_collection_elements,
                                        page=page,
                                        per_page=per_page)

    def archive(self, uid: Union[UUID, str], *, version: Union[int, str] = MOST_RECENT_VER):
        url = self._construct_path(uid, version, "archive")
        entity = self.session.put_resource(url, {}, version=self._api_version)
        return self.build(entity)

    def restore(self, uid: Union[UUID, str], *, version: Union[int, str] = MOST_RECENT_VER):
        url = self._construct_path(uid, version, "restore")
        entity = self.session.put_resource(url, {}, version=self._api_version)
        return self.build(entity)

    def convert_to_graph(self,
                         uid: Union[UUID, str],
                         *,
                         version: Union[int, str] = MOST_RECENT_VER,
                         retrain_if_needed: bool = False) -> Predictor:
        path = self._construct_path(uid, version, "convert")
        try:
            entity = self.session.get_resource(path, version=self._api_version)
        except Conflict as exc:
            if retrain_if_needed:
                self._train(uid, version)
                return None
            else:
                raise exc
        return self.build(entity)


class PredictorCollection(AbstractModuleCollection[Predictor]):
    """Represents the collection of all predictors for a project.

    Parameters
    ----------
    project_id: UUID
        the UUID of the project

    """

    _api_version = 'v3'
    _path_template = '/projects/{project_id}/predictors'
    _individual_key = None
    _resource = Predictor
    _collection_key = 'response'

    def __init__(self, project_id: UUID, session: Session):
        self.project_id = project_id
        self.session: Session = session
        self._versions_collection = _PredictorVersionCollection(project_id, session)

    def _predictors_path(self, subpath: str, uid: Union[UUID, str] = None) -> str:
        path_template = self._path_template + (f"/{uid}" if uid else "") + f"/{subpath}"
        return format_escaped_url(path_template, project_id=self.project_id)

    def build(self, data: dict) -> Predictor:
        """Build an individual Predictor."""
        predictor: Predictor = Predictor.build(data)
        predictor._session = self.session
        predictor._project_id = self.project_id
        return predictor

    def get(self,
            uid: Union[UUID, str],
            *,
            version: Union[int, str] = MOST_RECENT_VER) -> Predictor:
        """Get a predictor by ID and (optionally) version.

        If version is omitted, the most recent version will be retrieved.
        """
        if uid is None:
            raise ValueError("Cannot get when uid=None.  Are you using a registered resource?")
        return self._versions_collection.get(uid=uid, version=version)

    def register(self, predictor: Predictor) -> Predictor:
        """Register and train a Predictor.

        This predctor will be version 1, and its `draft` flag will be `True`. If training completes
        successfully, the `draft` flag will be set to `False`. Otherwise, it will remain `False`.
        """
        created_predictor = super().register(predictor)

        # If the initial response is invalid, just return it.
        # If not, kick off training since we never exposed saving a model without it,
        # so we should continue to do it automatically.
        if created_predictor.failed():
            return created_predictor
        else:
            return self._train(created_predictor.uid)

    def update(self, predictor: Predictor) -> Predictor:
        """Update and train a Predictor.

        If the predictor is a draft, this will overwrite its contents, then begin training. If it's
        not a draft, a new version will be created with the update, and then training will begin.

        In either case, if training completes successfully, it will no longer be a draft.
        """
        updated_predictor = super().update(predictor)

        # The /api/v3/predictors endpoint switched (un)archive from a field on the update payload
        # to their own endpoints. To maintain backwards compatibility, all predictors have an
        # _archived field set by the archived property. It will be archived if True, and restored
        # if False. It defaults to None, which does nothing. The value is reset afterwards.
        if predictor._archived is True:
            self.archive_root(predictor.uid)
        elif predictor._archived is False:
            self.restore_root(predictor.uid)
        predictor._archived = None

        # If the initial response is invalid, just return it
        # If not, kick off training since we never exposed saving a model without training
        # so we should continue to do it automatically
        if updated_predictor.failed():
            return updated_predictor
        else:
            return self._train(updated_predictor.uid)

    def _train(self, uid: Union[UUID, str]) -> Predictor:
        path = self._predictors_path("train", uid)
        params = {"create_version": True}
        entity = self.session.put_resource(path, {}, params=params, version=self._api_version)
        return self.build(entity)

    def archive_version(self, uid: Union[UUID, str], *, version: Union[int, str]) -> Predictor:
        """Archive a predictor version."""
        return self._versions_collection.archive(uid, version=version)

    def restore_version(self, uid: Union[UUID, str], *, version: Union[int, str]) -> Predictor:
        """Restore a predictor version."""
        return self._versions_collection.restore(uid, version=version)

    def archive_root(self, uid: Union[UUID, str]):
        """Archive a root predictor.

        uid: Union[UUID, str]
            Unique identifier of the predictor to archive.

        """
        path = self._predictors_path("archive", uid)
        self.session.put_resource(path, {}, version=self._api_version)

    def restore_root(self, uid: Union[UUID, str]):
        """Restore an archived root predictor.

        uid: Union[UUID, str]
            Unique identifier of the predictor to restore.

        """
        path = self._predictors_path("restore", uid)
        self.session.put_resource(path, {}, version=self._api_version)

    def root_is_archived(self, uid: Union[UUID, str]) -> bool:
        """Determine if the predictor root is archived.

        uid: Union[UUID, str]
            Unique identifier of the predictor to check.
        """
        uid = str(uid)
        return any(uid == str(archived_pred.uid) for archived_pred in self.list_archived())

    def list_versions(self,
                      uid: Union[UUID, str] = None,
                      *,
                      per_page: int = 100) -> Iterable[Predictor]:
        """List all non-archived versions of the given Predictor."""
        return self._versions_collection.list(uid, per_page=per_page)

    def list_archived(self,
                      *,
                      page: Optional[int] = None,
                      per_page: int = 20) -> Iterable[Predictor]:
        """List archived Predictors."""
        fetcher = partial(self._fetch_page, additional_params={"filter": "archived eq 'true'"})
        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_collection_elements,
                                        page=page,
                                        per_page=per_page)

    def list_archived_versions(self,
                               uid: Union[UUID, str] = None,
                               *,
                               per_page: int = 20) -> Iterable[Predictor]:
        """List all archived versions of the given Predictor."""
        return self._versions_collection.list_archived(uid, per_page=per_page)

    def check_for_update(self, uid: Union[UUID, str] = None,
                         predictor_id: Union[UUID, str] = None) -> Optional[Predictor]:
        """
        Check if there are updates available for a predictor.

        Typically these are updates to the training data. For example, a GEM Table may have
        been re-built to include additional rows.

        This check does not update the predictor; it just returns the update that is available.
        To perform the update, the response should then be used to call PredictorCollection.update

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the predictor to check
        predictor_id: Union[UUID, str]
            [DEPRECATED] please use uid instead

        Returns
        -------
        Optional[Predictor]
            The update, if an update is available; None otherwise.

        """
        uid = migrate_deprecated_argument(uid, "uid", predictor_id, "predictor_id")
        path = self._predictors_path("update-check", uid)
        update_data = self.session.get_resource(path, version=self._api_version)
        if update_data["updatable"]:
            built = Predictor.build(update_data)
            built.uid = uid
            return built
        else:
            return None

    def create_default(self,
                       *,
                       training_data: DataSource,
                       pattern: Union[str, AutoConfigureMode] = AutoConfigureMode.INFER,
                       prefer_valid: bool = True) -> Predictor:
        """[ALPHA] Create a default predictor for some training data.

        This method will return an unregistered predictor generated by inspecting the
        training data and attempting to automatically configure the predictor.

        The configuration generated while using the `AutoConfigureMode.SIMPLE` pattern
        includes featurizers for chemical formulas/molecular structures,
        and `AutoMLPredictor`s for any variables identified as responses in the training data.
        The configuration generated while using the `AutoConfigureMode.FORMULATION` pattern
        includes these same components, as well as a `SimpleMixturePredictor`,
        `LabelFractionsPredictor`, `IngredientFractionsPredictor`, and a series of
        `MeanPropertyPredictor`s to handle featurization of formulation quantities
        and ingredient properties.
        The `AutoConfigureMode.INFER` pattern chooses an appropriate mode based on whether
        the data source contains formulations data or not.

        Parameters
        ----------
        training_data: DataSource
            The data to configure the predictor to model.
        pattern: AutoConfigureMode or str
            The predictor pattern to use, either "PLAIN", "FORMULATION", or "INFER".
            The "INFER" pattern auto-detects whether the `DataSource` contains formulations
            data or not.
            If it does, then a formulation predictor is created.
            If not, then a plain predictor is created.
        prefer_valid: Boolean
            If True, enables filtering of sparse descriptors and trimming of
            excess graph components in attempt to return a default configuration
            that will pass validation.
            Default: True.

        Returns
        -------
        Predictor
            Automatically configured predictor for the training data

        """
        # Continue handling string pattern inputs
        if not isinstance(pattern, AutoConfigureMode):
            pattern = AutoConfigureMode.get_enum(pattern)
        pattern = pattern.value

        path = self._predictors_path("default")
        body = {"data_source": training_data.dump(), "pattern": pattern,
                "prefer_valid": prefer_valid}
        data = self.session.post_resource(path, json=body, version=self._api_version)
        return self.build(Predictor.wrap_instance(data["instance"]))

    def convert_to_graph(self,
                         uid: Union[UUID, str],
                         retrain_if_needed: bool = False,
                         *,
                         version: Union[int, str] = MOST_RECENT_VER) -> Predictor:
        """Given a SimpleML or Graph predictor, get an equivalent Graph predictor.

        Returns a Graph predictor with any SimpleML predictors converted to an equivalent AutoML
        predictor. If it's not a SimpleML or Graph predictor, or it's not in the READY state, an
        error is raised. SimpleML predictors are deprecated, so this is to aid in your migration.

        Note this conversion is not performed in place! That is, the predictor returned is not
        persisted on the platform. To persist it, pass the converted predictor to
        :py:meth:`citrine.resources.project.PredictorCollection.update`. Or you can do this in
        one step with :py:meth:`citrine.resources.project.PredictorCollection.convert_and_update`.

        .. code:: python

            converted = project.predictors.convert_to_graph(predictor_uid)
            project.predictors.update(converted)

            # is equivalent to

            converted = project.predictors.convert_and_update(predictor_uid)

        If a predictor needs to be retrained before conversion, it will raise an HTTP 409 Conflict
        error. This may occur when the Citrine Platform has been updated since your predictor was
        last trained. If `retrain_if_needed` is `True`, retraining will automatically begin, and
        the method will return `None. Once retraining completes, call this method again to get the
        converted predictor. For example:

        .. code:: python

            converted = project.predictors.convert_and_update(pred.uid, retrain_if_needed=True)
            if converted is None:
                predictor = project.predictors.get(pred.uid)
                wait_while_validating(collection=project.predictors, module=predictor)
                converted = project.predictors.convert_and_update(pred.uid)
        """
        return self._versions_collection.convert_to_graph(uid,
                                                          version=version,
                                                          retrain_if_needed=retrain_if_needed)

    def convert_and_update(self,
                           uid: Union[UUID, str],
                           retrain_if_needed: bool = False,
                           *,
                           version: Union[int, str] = MOST_RECENT_VER) -> Predictor:
        """Given a SimpleML or Graph predictor, overwrite it with an equivalent Graph predictor.

        See `PredictorCollection.convert_to_graph` for more detail.
        """
        new_pred = self.convert_to_graph(uid, version=version, retrain_if_needed=retrain_if_needed)
        return self.update(new_pred) if new_pred else None

    @deprecated(deprecated_in="1.50.0", removed_in="2.0.0",
                details="Use PredictorCollection.get() instead.")
    def get_version(self, uid: Union[UUID, str], *, version: Union[int, str]) -> Predictor:
        """[DEPRECATED] Get a specific version of the predictor."""
        return self.get(uid=uid, version=version)

    @deprecated(deprecated_in="1.39.0", removed_in="2.0.0",
                details="archive() is deprecated in favor of archive_root().")
    def archive(self, uid: Union[UUID, str]) -> Predictor:
        """[DEPRECATED] Archive a root predictor."""
        self.archive_root(uid)
        return self.get(uid)

    @deprecated(deprecated_in="1.39.0", removed_in="2.0.0",
                details="restore() is deprecated in favor of restore_root().")
    def restore(self, uid: Union[UUID, str]) -> Predictor:
        """[DEPRECATED] Restore a root predictor."""
        self.restore_root(uid)
        return self.get(uid)

    @deprecated(deprecated_in="1.47.0", removed_in="2.0.0",
                details="auto_configure is an alias for create_default.")
    def auto_configure(self,
                       *,
                       training_data: DataSource,
                       pattern: Union[str, AutoConfigureMode] = AutoConfigureMode.INFER,
                       prefer_valid: bool = True) -> Predictor:
        """[DEPRECATED] Alias for PredictorCollection.create_default."""
        return self.create_default(training_data=training_data, pattern=pattern,
                                   prefer_valid=prefer_valid)

    @deprecated(deprecated_in="1.50.0", removed_in="2.0.0",
                details="Use PredictorCollection.convert_to_graph() instead.")
    def convert_version_to_graph(self,
                                 uid: Union[UUID, str],
                                 *,
                                 version: Union[int, str],
                                 retrain_if_needed: bool = False) -> Predictor:
        """[DEPRECATED] Given a SimpleML or Graph predictor, get an equivalent Graph predictor.

        See `PredictorCollection.convert_to_graph` for more detail.
        """
        return self.convert_to_graph(uid, version=version, retrain_if_needed=retrain_if_needed)

    @deprecated(deprecated_in="1.50.0", removed_in="2.0.0",
                details="Use PredictorCollection.convert_and_update() instead.")
    def convert_version_and_update(self,
                                   uid: Union[UUID, str],
                                   *,
                                   version: Union[int, str],
                                   retrain_if_needed: bool = False) -> Predictor:
        """[DEPRECATED] Overwrite a SimpleML or Graph predictor with an equivalent Graph predictor.

        See `PredictorCollection.convert_to_graph` for more detail.
        """
        return self.convert_and_update(uid, version=version, retrain_if_needed=retrain_if_needed)
